@@ -158,7 +158,6 @@ struct kgem {
 	uint16_t nreloc__self;
 	uint16_t nfence;
 	uint16_t batch_size;
-	uint16_t min_alignment;
 
 	uint32_t flush:1;
 	uint32_t need_expire:1;
@@ -339,6 +338,7 @@ static inline void kgem_submit(struct kgem *kgem)
 
 static inline void kgem_bo_submit(struct kgem *kgem, struct kgem_bo *bo)
 {
+	assert(bo->refcnt);
 	if (bo->exec)
 		_kgem_submit(kgem);
 }
@@ -515,6 +515,8 @@ static inline bool kgem_bo_blt_pitch_is_ok(struct kgem *kgem,
 static inline bool kgem_bo_can_blt(struct kgem *kgem,
 				   struct kgem_bo *bo)
 {
+	assert(bo->refcnt);
+
 	if (bo->tiling == I915_TILING_Y) {
 		DBG(("%s: can not blt to handle=%d, tiling=Y\n",
 		     __FUNCTION__, bo->handle));
@@ -538,6 +540,7 @@ bool __kgem_busy(struct kgem *kgem, int handle);
 
 static inline void kgem_bo_mark_busy(struct kgem_bo *bo, int ring)
 {
+	assert(bo->refcnt);
 	bo->rq = (struct kgem_request *)((uintptr_t)bo->rq | ring);
 }
 
@@ -584,9 +587,11 @@ static inline bool kgem_bo_is_render(struct kgem_bo *bo)
 
 static inline void kgem_bo_mark_unreusable(struct kgem_bo *bo)
 {
+	assert(bo->refcnt);
 	while (bo->proxy) {
 		bo->flush = true;
 		bo = bo->proxy;
+		assert(bo->refcnt);
 	}
 	bo->flush = true;
 	bo->reusable = false;
@@ -617,6 +622,8 @@ static inline void __kgem_bo_mark_dirty(struct kgem_bo *bo)
 	DBG(("%s: handle=%d (proxy? %d)\n", __FUNCTION__,
 	     bo->handle, bo->proxy != NULL));
 
+	assert(bo->refcnt);
+
 	bo->exec->flags |= LOCAL_EXEC_OBJECT_WRITE;
 	bo->needs_flush = bo->gpu_dirty = true;
 	list_move(&bo->request, &RQ(bo->rq)->buffers);
@@ -661,16 +668,15 @@ static inline bool kgem_bo_can_map(struct kgem *kgem, struct kgem_bo *bo)
 	if (kgem->gen == 021 && bo->tiling == I915_TILING_Y)
 		return false;
 
-	if (!bo->presumed_offset)
-		return __kgem_bo_num_pages(bo) <= kgem->aperture_mappable / 4;
-
-	return bo->presumed_offset / PAGE_SIZE + __kgem_bo_num_pages(bo) <= kgem->aperture_mappable;
+	return __kgem_bo_num_pages(bo) <= kgem->aperture_mappable / 4;
 }
 
 static inline bool kgem_bo_can_map__cpu(struct kgem *kgem,
 					struct kgem_bo *bo,
 					bool write)
 {
+	assert(bo->refcnt);
+
 	if (bo->purged || (bo->scanout && write))
 		return false;
 
