@@ -57,7 +57,6 @@ USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include <xf86RandR12.h>
 #include <mi.h>
 #include <micmap.h>
-#include <mipict.h>
 
 #include <sys/ioctl.h>
 #include <sys/fcntl.h>
@@ -83,8 +82,7 @@ sna_load_palette(ScrnInfoPtr scrn, int numColors, int *indices,
 		 LOCO * colors, VisualPtr pVisual)
 {
 	xf86CrtcConfigPtr xf86_config = XF86_CRTC_CONFIG_PTR(scrn);
-	int i, j, index;
-	int p;
+	int p, n, i, j;
 	uint16_t lut_r[256], lut_g[256], lut_b[256];
 
 	DBG(("%s\n", __FUNCTION__));
@@ -92,48 +90,43 @@ sna_load_palette(ScrnInfoPtr scrn, int numColors, int *indices,
 	for (p = 0; p < xf86_config->num_crtc; p++) {
 		xf86CrtcPtr crtc = xf86_config->crtc[p];
 
+#define C(I,RGB) (colors[I].RGB << 8 | colors[I].RGB)
 		switch (scrn->depth) {
 		case 15:
-			for (i = 0; i < numColors; i++) {
-				index = indices[i];
+			for (n = 0; n < numColors; n++) {
+				i = indices[n];
 				for (j = 0; j < 8; j++) {
-					lut_r[index * 8 + j] =
-					    colors[index].red << 8;
-					lut_g[index * 8 + j] =
-					    colors[index].green << 8;
-					lut_b[index * 8 + j] =
-					    colors[index].blue << 8;
+					lut_r[8*i + j] = C(i, red);
+					lut_g[8*i + j] = C(i, green);
+					lut_b[8*i + j] = C(i, blue);
 				}
 			}
 			break;
 		case 16:
-			for (i = 0; i < numColors; i++) {
-				index = indices[i];
+			for (n = 0; n < numColors; n++) {
+				i = indices[n];
 
-				if (index <= 31) {
+				if (i <= 31) {
 					for (j = 0; j < 8; j++) {
-						lut_r[index * 8 + j] =
-						    colors[index].red << 8;
-						lut_b[index * 8 + j] =
-						    colors[index].blue << 8;
+						lut_r[8*i + j] = C(i, red);
+						lut_b[8*i + j] = C(i, blue);
 					}
 				}
 
-				for (j = 0; j < 4; j++) {
-					lut_g[index * 4 + j] =
-					    colors[index].green << 8;
-				}
+				for (j = 0; j < 4; j++)
+					lut_g[4*i + j] = C(i, green);
 			}
 			break;
 		default:
-			for (i = 0; i < numColors; i++) {
-				index = indices[i];
-				lut_r[index] = colors[index].red << 8;
-				lut_g[index] = colors[index].green << 8;
-				lut_b[index] = colors[index].blue << 8;
+			for (n = 0; n < numColors; n++) {
+				i = indices[n];
+				lut_r[i] = C(i, red);
+				lut_g[i] = C(i, green);
+				lut_b[i] = C(i, blue);
 			}
 			break;
 		}
+#undef C
 
 		/* Make the change through RandR */
 #ifdef RANDR_12_INTERFACE
@@ -411,7 +404,6 @@ static Bool fb_supports_depth(int fd, int depth)
 {
 	struct drm_i915_gem_create create;
 	struct drm_mode_fb_cmd fb;
-	struct drm_gem_close close;
 	Bool ret;
 
 	VG_CLEAR(create);
@@ -431,9 +423,7 @@ static Bool fb_supports_depth(int fd, int depth)
 	ret = drmIoctl(fd, DRM_IOCTL_MODE_ADDFB, &fb) == 0;
 	drmModeRmFB(fd, fb.fb_id);
 
-	VG_CLEAR(close);
-	close.handle = create.handle;
-	(void)drmIoctl(fd, DRM_IOCTL_GEM_CLOSE, &close);
+	(void)drmIoctl(fd, DRM_IOCTL_GEM_CLOSE, &create.handle);
 
 	return ret;
 }
@@ -501,8 +491,6 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	scrn->displayWidth = 640;	/* default it */
 
-	sna->PciInfo = xf86GetPciInfoForEntity(sna->pEnt->index);
-
 	scrn->monitor = scrn->confScreen->monitor;
 	scrn->progClock = TRUE;
 	scrn->rgbBits = 8;
@@ -521,6 +509,10 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 		goto cleanup;
 	}
 
+	intel_detect_chipset(scrn, pEnt);
+	xf86DrvMsg(scrn->scrnIndex, X_PROBED, "CPU: %s\n",
+		   sna_cpu_features_to_string(sna->cpu_features, buf));
+
 	preferred_depth = sna->info->gen < 030 ? 15 : 24;
 	if (!fb_supports_depth(fd, preferred_depth))
 		preferred_depth = 24;
@@ -558,9 +550,9 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	sna_setup_capabilities(scrn, fd);
 
-	intel_detect_chipset(scrn, sna->pEnt, sna->PciInfo);
-
-	kgem_init(&sna->kgem, fd, sna->PciInfo, sna->info->gen);
+	kgem_init(&sna->kgem, fd,
+		  xf86GetPciInfoForEntity(pEnt->index),
+		  sna->info->gen);
 	if (xf86ReturnOptValBool(sna->Options, OPTION_ACCEL_DISABLE, FALSE) ||
 	    !sna_option_cast_to_bool(sna, OPTION_ACCEL_METHOD, TRUE)) {
 		xf86DrvMsg(sna->scrn->scrnIndex, X_CONFIG,
@@ -603,8 +595,6 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 	if (xf86ReturnOptValBool(sna->Options, OPTION_CRTC_PIXMAPS, FALSE))
 		sna->flags |= SNA_FORCE_SHADOW;
 
-	xf86DrvMsg(scrn->scrnIndex, X_PROBED, "CPU: %s\n",
-		   sna_cpu_features_to_string(sna->cpu_features, buf));
 	xf86DrvMsg(scrn->scrnIndex, X_CONFIG, "Framebuffer %s\n",
 		   sna->tiling & SNA_TILING_FB ? "tiled" : "linear");
 	xf86DrvMsg(scrn->scrnIndex, X_CONFIG, "Pixmaps %s\n",
@@ -776,10 +766,8 @@ sna_uevent_init(ScrnInfoPtr scrn)
 		return;
 	}
 
-	sna->uevent_handler =
-		xf86AddGeneralHandler(udev_monitor_get_fd(mon),
-				      sna_handle_uevents,
-				      scrn);
+	sna->uevent_handler = xf86AddGeneralHandler(udev_monitor_get_fd(mon),
+						    sna_handle_uevents, scrn);
 	if (!sna->uevent_handler) {
 		udev_monitor_unref(mon);
 		udev_unref(u);
@@ -939,6 +927,7 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 {
 	ScrnInfoPtr scrn = xf86ScreenToScrn(screen);
 	struct sna *sna = to_sna(scrn);
+	struct pci_device *pci;
 	VisualPtr visuals;
 	DepthPtr depths;
 	int nvisuals;
@@ -956,7 +945,11 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 	if (!sna_register_all_privates())
 		return FALSE;
 
-	scrn->videoRam = agp_aperture_size(sna->PciInfo, sna->kgem.gen) / 1024;
+	pci = xf86GetPciInfoForEntity(sna->pEnt->index);
+	if (pci != NULL)
+		scrn->videoRam = agp_aperture_size(pci, sna->kgem.gen) / 1024;
+	else
+		scrn->videoRam = 256;
 
 	miClearVisualTypes();
 	if (!miSetVisualTypes(scrn->depth,
@@ -1041,9 +1034,7 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 	if (!xf86CrtcScreenInit(screen))
 		return FALSE;
 
-	xf86RandR12SetRotations(screen,
-				RR_Rotate_0 | RR_Rotate_90 | RR_Rotate_180 | RR_Rotate_270 |
-				RR_Reflect_X | RR_Reflect_Y);
+	xf86RandR12SetRotations(screen, RR_Rotate_All | RR_Reflect_All);
 	xf86RandR12SetTransformSupport(screen, TRUE);
 
 	if (!miCreateDefColormap(screen))
