@@ -144,6 +144,10 @@ struct sna_composite_op {
 		struct {
 			uint32_t flags;
 		} gen7;
+
+		struct {
+			uint32_t flags;
+		} gen8;
 	} u;
 
 	void *priv;
@@ -503,6 +507,57 @@ struct gen7_render_state {
 	bool emit_flush;
 };
 
+
+enum {
+	GEN8_WM_KERNEL_NOMASK = 0,
+	GEN8_WM_KERNEL_NOMASK_P,
+
+	GEN8_WM_KERNEL_MASK,
+	GEN8_WM_KERNEL_MASK_P,
+
+	GEN8_WM_KERNEL_MASKCA,
+	GEN8_WM_KERNEL_MASKCA_P,
+
+	GEN8_WM_KERNEL_MASKSA,
+	GEN8_WM_KERNEL_MASKSA_P,
+
+	GEN8_WM_KERNEL_OPACITY,
+	GEN8_WM_KERNEL_OPACITY_P,
+
+	GEN8_WM_KERNEL_VIDEO_PLANAR,
+	GEN8_WM_KERNEL_VIDEO_PACKED,
+	GEN8_WM_KERNEL_COUNT
+};
+
+struct gen8_render_state {
+	unsigned gt;
+
+	struct kgem_bo *general_bo;
+
+	uint32_t vs_state;
+	uint32_t sf_state;
+	uint32_t sf_mask_state;
+	uint32_t wm_state;
+	uint32_t wm_kernel[GEN8_WM_KERNEL_COUNT][3];
+
+	uint32_t cc_blend;
+
+	uint32_t drawrect_offset;
+	uint32_t drawrect_limit;
+	uint32_t blend;
+	uint32_t samplers;
+	uint32_t kernel;
+
+	uint16_t num_sf_outputs;
+	uint16_t ve_id;
+	uint16_t last_primitive;
+	int16_t floats_per_vertex;
+	uint16_t surface_table;
+
+	bool needs_invariant;
+	bool emit_flush;
+};
+
 struct sna_static_stream {
 	uint32_t size, used;
 	uint8_t *data;
@@ -557,6 +612,7 @@ const char *gen4_render_init(struct sna *sna, const char *backend);
 const char *gen5_render_init(struct sna *sna, const char *backend);
 const char *gen6_render_init(struct sna *sna, const char *backend);
 const char *gen7_render_init(struct sna *sna, const char *backend);
+const char *gen8_render_init(struct sna *sna, const char *backend);
 
 bool sna_tiling_composite(uint32_t op,
 			  PicturePtr src,
