@@ -36,6 +36,8 @@
 #include <X11/extensions/shmproto.h>
 #elif HAVE_X11_EXTENSIONS_SHMSTR_H
 #include <X11/extensions/shmstr.h>
+#else
+#error Failed to find the right header for X11 MIT-SHM protocol definitions
 #endif
 #include <X11/extensions/Xdamage.h>
 #include <X11/extensions/Xinerama.h>
@@ -59,7 +61,6 @@
 #include <signal.h>
 #include <getopt.h>
 #include <limits.h>
-#include <ctype.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <assert.h>
@@ -84,7 +85,10 @@ struct display {
 	Window root;
 	Visual *visual;
 	Damage damage;
+	long timestamp;
 
+	int width;
+	int height;
 	int depth;
 
 	XRenderPictFormat *root_format;
@@ -109,6 +113,8 @@ struct display {
 
 	int flush;
 	int send;
+	int skip_clone;
+	int skip_frame;
 };
 
 struct output {
@@ -140,6 +146,7 @@ struct clone {
 	struct clone *active;
 
 	struct output src, dst;
+	long timestamp;
 
 	XShmSegmentInfo shm;
 	XImage image;
@@ -258,6 +265,8 @@ can_use_shm(Display *dpy,
 	if (codes) {
 		XShmCompletionEvent e;
 
+		memset(&e, 0, sizeof(e));
+
 		e.type = codes->first_event;
 		e.send_event = 1;
 		e.serial = 1;
@@ -346,7 +355,9 @@ static void context_enable_timer(struct context *ctx)
 	if (ctx->timer_active)
 		return;
 
-	read(ctx->timer, &count, sizeof(count));
+	/* reset timer */
+	count = read(ctx->timer, &count, sizeof(count));
+
 	ctx->timer_active = 1;
 }
 
@@ -408,6 +419,23 @@ static XRRModeInfo *lookup_mode(XRRScreenResources *res, int id)
 	return NULL;
 }
 
+static void clone_update_edid(struct clone *clone)
+{
+	unsigned long nitems, after;
+	unsigned char *data;
+	int format;
+	Atom type;
+
+	if (XRRGetOutputProperty(clone->dst.dpy, clone->dst.rr_output,
+				 XInternAtom(clone->dst.dpy, "EDID", False),
+				 0, 100, False, False, AnyPropertyType,
+				 &type, &format, &nitems, &after, &data) == Success) {
+		XRRChangeOutputProperty(clone->src.dpy, clone->src.rr_output,
+					XInternAtom(clone->src.dpy, "EDID", False),
+					type, format, PropModeReplace, data, nitems);
+	}
+}
+
 static int clone_update_modes__randr(struct clone *clone)
 {
 	XRRScreenResources *from_res = NULL, *to_res = NULL;
@@ -426,6 +454,12 @@ static int clone_update_modes__randr(struct clone *clone)
 	if (from_info == NULL)
 		goto err;
 
+	DBG(("%s(%s-%s): timestamp %ld (last %ld)\n", __func__,
+	     DisplayString(clone->src.dpy), clone->src.name,
+	    from_info->timestamp, clone->timestamp));
+	if (from_info->timestamp == clone->timestamp)
+		goto err;
+
 	to_res = _XRRGetScreenResourcesCurrent(clone->src.dpy, clone->src.window);
 	if (to_res == NULL)
 		goto err;
@@ -434,6 +468,30 @@ static int clone_update_modes__randr(struct clone *clone)
 	if (to_info == NULL)
 		goto err;
 
+	if (clone->dst.rr_crtc == from_info->crtc) {
+		for (i = 0; i < to_info->nmode; i++) {
+			XRRModeInfo *mode, *old;
+
+			mode = lookup_mode(to_res, to_info->modes[i]);
+			if (mode == NULL)
+				break;
+			for (j = 0; j < from_info->nmode; j++) {
+				old = lookup_mode(from_res, from_info->modes[j]);
+				if (old && mode_equal(mode, old)) {
+					mode = NULL;
+					break;
+				}
+			}
+			if (mode)
+				break;
+		}
+		if (i == from_info->nmode && i == to_info->nmode) {
+			DBG(("%s(%s-%s): no change in output\n", __func__,
+			     DisplayString(clone->src.dpy), clone->src.name));
+			goto done;
+		}
+	}
+
 	clone->dst.rr_crtc = from_info->crtc;
 
 	/* Clear all current UserModes on the output, including any active ones */
@@ -489,12 +547,17 @@ static int clone_update_modes__randr(struct clone *clone)
 			m.name = buf;
 
 			id = XRRCreateMode(clone->src.dpy, clone->src.window, &m);
+			DBG(("%s(%s-%s): adding mode %ld: %s\n", __func__,
+			     DisplayString(clone->src.dpy), clone->src.name, id, mode->name));
 		}
 
 		XRRAddOutputMode(clone->src.dpy, clone->src.rr_output, id);
 	}
+	clone_update_edid(clone);
 	XUngrabServer(clone->src.dpy);
+done:
 	ret = 0;
+	clone->timestamp = from_info->timestamp;
 
 err:
 	if (to_info)
@@ -691,6 +754,35 @@ static void init_image(struct clone *clone)
 
 	ret = XInitImage(image);
 	assert(ret);
+	(void)ret;
+}
+
+static int mode_height(const XRRModeInfo *mode, Rotation rotation)
+{
+	switch (rotation & 0xf) {
+	case RR_Rotate_0:
+	case RR_Rotate_180:
+		return mode->height;
+	case RR_Rotate_90:
+	case RR_Rotate_270:
+		return mode->width;
+	default:
+		return 0;
+	}
+}
+
+static int mode_width(const XRRModeInfo *mode, Rotation rotation)
+{
+	switch (rotation & 0xf) {
+	case RR_Rotate_0:
+	case RR_Rotate_180:
+		return mode->width;
+	case RR_Rotate_90:
+	case RR_Rotate_270:
+		return mode->height;
+	default:
+		return 0;
+	}
 }
 
 static void output_init_xfer(struct clone *clone, struct output *output)
@@ -732,6 +824,8 @@ static void output_init_xfer(struct clone *clone, struct output *output)
 
 static int clone_init_xfer(struct clone *clone)
 {
+	int width, height;
+
 	if (clone->src.mode.id == 0) {
 		if (clone->width == 0 && clone->height == 0)
 			return 0;
@@ -749,24 +843,29 @@ static int clone_init_xfer(struct clone *clone)
 			clone->shm.shmaddr = 0;
 		}
 
+		clone->damaged.x2 = clone->damaged.y2 = INT_MIN;
+		clone->damaged.x1 = clone->damaged.y1 = INT_MAX;
 		return 0;
 	}
 
-	if (clone->src.mode.width == clone->width &&
-	    clone->src.mode.height == clone->height)
+	width = mode_width(&clone->src.mode, clone->src.rotation);
+	height = mode_height(&clone->src.mode, clone->src.rotation);
+
+	if (width == clone->width && height == clone->height)
 		return 0;
 
-	DBG(("%s-%s create xfer\n",
-	     DisplayString(clone->dst.dpy), clone->dst.name));
+	DBG(("%s-%s create xfer, %dx%d\n",
+	     DisplayString(clone->dst.dpy), clone->dst.name,
+	     width, height));
 
-	clone->width = clone->src.mode.width;
-	clone->height = clone->src.mode.height;
+	clone->width = width;
+	clone->height = height;
 
 	if (clone->shm.shmaddr)
 		shmdt(clone->shm.shmaddr);
 
 	clone->shm.shmid = shmget(IPC_PRIVATE,
-				  clone->height * stride_for_depth(clone->width, clone->depth),
+				  height * stride_for_depth(width, clone->depth),
 				  IPC_CREAT | 0666);
 	if (clone->shm.shmid == -1)
 		return errno;
@@ -798,9 +897,9 @@ static int clone_init_xfer(struct clone *clone)
 	output_init_xfer(clone, &clone->dst);
 
 	clone->damaged.x1 = clone->src.x;
-	clone->damaged.x2 = clone->src.x + clone->width;
+	clone->damaged.x2 = clone->src.x + width;
 	clone->damaged.y1 = clone->src.y;
-	clone->damaged.y2 = clone->src.y + clone->height;
+	clone->damaged.y2 = clone->src.y + height;
 
 	display_mark_flush(clone->dst.display);
 	return 0;
@@ -818,7 +917,7 @@ static void clone_update(struct clone *clone)
 	clone->rr_update = 0;
 }
 
-static void context_update(struct context *ctx)
+static int context_update(struct context *ctx)
 {
 	Display *dpy = ctx->display->dpy;
 	XRRScreenResources *res;
@@ -829,8 +928,16 @@ static void context_update(struct context *ctx)
 
 	res = _XRRGetScreenResourcesCurrent(dpy, ctx->display->root);
 	if (res == NULL)
-		return;
+		return 0;
 
+	DBG(("%s timestamp %ld (last %ld)\n", DisplayString(dpy), res->timestamp, ctx->display->timestamp));
+	if (res->timestamp == ctx->display->timestamp &&
+	    res->timestamp != res->configTimestamp) { /* mutter be damned */
+		XRRFreeScreenResources(res);
+		return 0;
+	}
+
+	ctx->display->timestamp = res->timestamp;
 	for (n = 0; n < ctx->nclone; n++) {
 		struct output *output = &ctx->clones[n].src;
 		XRROutputInfo *o;
@@ -849,7 +956,7 @@ static void context_update(struct context *ctx)
 			DBG(("%s-%s: (x=%d, y=%d, rotation=%d, mode=%ld) -> (x=%d, y=%d, rotation=%d, mode=%ld)\n",
 			     DisplayString(dpy), output->name,
 			     output->x, output->y, output->rotation, output->mode.id,
-			     c->x, c->y, output->rotation, c->mode));
+			     c->x, c->y, c->rotation, c->mode));
 
 			changed |= output->rotation |= c->rotation;
 			output->rotation = c->rotation;
@@ -896,7 +1003,7 @@ static void context_update(struct context *ctx)
 
 	DBG(("%s changed? %d\n", DisplayString(dpy), context_changed));
 	if (!context_changed)
-		return;
+		return 0;
 
 	for (n = 1; n < ctx->ndisplay; n++) {
 		struct display *display = &ctx->display[n];
@@ -921,35 +1028,68 @@ static void context_update(struct context *ctx)
 			DBG(("%s: source %s enabled (%d, %d)x(%d, %d)\n",
 			     DisplayString(clone->dst.dpy), output->name,
 			     output->x, output->y,
-			     output->mode.width, output->mode.height));
+			     mode_width(&output->mode, output->rotation),
+			     mode_height(&output->mode, output->rotation)));
 
 			if (output->x < x1)
 				x1 = output->x;
 			if (output->y < y1)
 				y1 = output->y;
 
-			v = (int)output->x + output->mode.width;
+			v = (int)output->x + mode_width(&output->mode, output->rotation);
 			if (v > x2)
 				x2 = v;
-			v = (int)output->y + output->mode.height;
+			v = (int)output->y + mode_height(&output->mode, output->rotation);
 			if (v > y2)
 				y2 = v;
 		}
 
+		x2 -= x1;
+		y2 -= y1;
 		DBG(("%s fb bounds (%d, %d)x(%d, %d)\n", DisplayString(display->dpy),
-		     x1, y1, x2-x1, y2-y1));
+		     x1, y1, x2, y2));
 
 		res = _XRRGetScreenResourcesCurrent(display->dpy, display->root);
 		if (res == NULL)
 			continue;
 
 		XGrabServer(display->dpy);
+
+		DBG(("%s: current size %dx%d, need %dx%d\n",
+		     DisplayString(display->dpy),
+		     display->width, display->height,
+		     x2, y2));
+
+		if (display->width != x2 || display->height != y2) {
+			/* When shrinking we have to manually resize the fb */
+			for (clone = display->clone; clone; clone = clone->next) {
+				struct output *dst = &clone->dst;
+
+				if (!dst->rr_crtc)
+					continue;
+
+				DBG(("%s: disabling output '%s'\n",
+				     DisplayString(dst->dpy), dst->name));
+				XRRSetCrtcConfig(dst->dpy, res, dst->rr_crtc, CurrentTime,
+						0, 0, None, RR_Rotate_0, NULL, 0);
+				dst->rr_crtc = 0;
+				dst->mode.id = 0;
+			}
+
+			DBG(("%s: XRRSetScreenSize %dx%d\n", DisplayString(display->dpy), x2, y2));
+			XRRSetScreenSize(display->dpy, display->root, x2, y2, x2 * 96 / 25.4, y2 * 96 / 25.4);
+			display->width = x2;
+			display->height = y2;
+		}
+
 		for (clone = display->clone; clone; clone = clone->next) {
 			struct output *src = &clone->src;
 			struct output *dst = &clone->dst;
 			XRROutputInfo *o;
+			XRRPanning panning;
 			struct clone *set;
 			RRCrtc rr_crtc;
+			Status ret;
 
 			DBG(("%s: copying configuration from %s (mode=%ld) to %s\n",
 			     DisplayString(dst->dpy), src->name, (long)src->mode.id, dst->name));
@@ -1016,18 +1156,26 @@ err:
 				XRRFreeOutputInfo(o);
 			}
 			if (rr_crtc == 0) {
-				DBG(("%s: failed to find availble CRTC for %s\n",
+				DBG(("%s: failed to find available CRTC for %s\n",
 				     DisplayString(dst->dpy), dst->name));
 				goto err;
 			}
 
-			DBG(("%s: enabling output '%s' (%d,%d)x(%d,%d) on CRTC:%ld\n",
+			DBG(("%s: enabling output '%s' (%d,%d)x(%d,%d), rotation %d, on CRTC:%ld\n",
 			     DisplayString(dst->dpy), dst->name,
-			     dst->x, dst->y, dst->mode.width, dst->mode.height, (long)rr_crtc));
-			XRRSetCrtcConfig(dst->dpy, res, rr_crtc, CurrentTime,
-					 dst->x, dst->y, dst->mode.id, dst->rotation,
-					 &dst->rr_output, 1);
+			     dst->x, dst->y, dst->mode.width, dst->mode.height,
+			     dst->rotation, (long)rr_crtc));
+
+			ret = XRRSetCrtcConfig(dst->dpy, res, rr_crtc, CurrentTime,
+					       dst->x, dst->y, dst->mode.id, dst->rotation,
+					       &dst->rr_output, 1);
+			DBG(("%s-%s: XRRSetCrtcConfig %s\n", DisplayString(dst->dpy), dst->name, ret ? "failed" : "success"));
+
+			ret = XRRSetPanning(dst->dpy, res, rr_crtc, memset(&panning, 0, sizeof(panning)));
+			DBG(("%s-%s: XRRSetPanning %s\n", DisplayString(dst->dpy), dst->name, ret ? "failed" : "success"));
+
 			dst->rr_crtc = rr_crtc;
+			(void)ret;
 		}
 		XUngrabServer(display->dpy);
 
@@ -1044,6 +1192,8 @@ err:
 		clone->active = ctx->active;
 		ctx->active = clone;
 	}
+
+	return 1;
 }
 
 static Cursor display_load_invisible_cursor(struct display *display)
@@ -1132,6 +1282,8 @@ static void display_flush_cursor(struct display *display)
 		y = display->cursor_y++ & 31;
 	}
 
+	DBG(("%s setting cursor position (%d, %d), visible? %d\n",
+	     DisplayString(display->dpy), x, y, display->cursor_visible));
 	XWarpPointer(display->dpy, None, display->root, 0, 0, 0, 0, x, y);
 
 	cursor = None;
@@ -1351,8 +1503,22 @@ static int clone_paint(struct clone *c)
 	DBG(("%s-%s is damaged, last SHM serial: %ld, now %ld\n",
 	     DisplayString(c->dst.dpy), c->dst.name,
 	     (long)c->dst.serial, (long)LastKnownRequestProcessed(c->dst.dpy)));
-	if (c->dst.serial > LastKnownRequestProcessed(c->dst.dpy))
-		return EAGAIN;
+	if (c->dst.serial > LastKnownRequestProcessed(c->dst.dpy)) {
+		struct pollfd pfd;
+
+		pfd.fd = ConnectionNumber(c->dst.dpy);
+		pfd.events = POLLIN;
+		XEventsQueued(c->dst.dpy,
+			      poll(&pfd, 1, 0) ? QueuedAfterReading : QueuedAfterFlush);
+
+		if (c->dst.serial > LastKnownRequestProcessed(c->dst.dpy)) {
+			c->dst.display->skip_clone++;
+			return EAGAIN;
+		}
+	}
+
+	c->dst.display->skip_clone = 0;
+	c->dst.display->skip_frame = 0;
 
 	if (FORCE_FULL_REDRAW) {
 		c->damaged.x1 = c->src.x;
@@ -1454,16 +1620,29 @@ static int record_mouse(struct context *ctx)
 
 static int bad_visual(Visual *visual, int depth)
 {
+	DBG(("%s? depth=%d, visual: class=%d, bits_per_rgb=%d, red_mask=%08lx, green_mask=%08lx, blue_mask=%08lx\n",
+	     __func__, depth,
+	     visual->class,
+	     visual->bits_per_rgb,
+	     visual->red_mask,
+	     visual->green_mask,
+	     visual->blue_mask));
+
+	if (!(visual->class == TrueColor || visual->class == DirectColor))
+		return 1;
+
 	switch (depth) {
-	case 16: return (visual->bits_per_rgb != 6 ||
-			 visual->red_mask != 0x1f << 11 ||
-			 visual->green_mask != 0x3f << 5 ||
-			 visual->blue_mask != 0x1f << 0);
-	case 24: return (visual->bits_per_rgb != 8 ||
-			 visual->red_mask != 0xff << 16 ||
-			 visual->green_mask != 0xff << 8 ||
-			 visual->blue_mask != 0xff << 0);
-	default: return 0;
+	case 16: return (/* visual->bits_per_rgb != 6          || */
+			 visual->red_mask     != 0x1f << 11 ||
+			 visual->green_mask   != 0x3f << 5  ||
+			 visual->blue_mask    != 0x1f << 0);
+
+	case 24: return (/* visual->bits_per_rgb != 8          || */
+			 visual->red_mask     != 0xff << 16 ||
+			 visual->green_mask   != 0xff << 8  ||
+			 visual->blue_mask    != 0xff << 0);
+
+	default: return 1;
 	}
 }
 
@@ -1654,6 +1833,7 @@ static int clone_init_depth(struct clone *clone)
 static int add_display(struct context *ctx, Display *dpy)
 {
 	struct display *display;
+	int first_display = ctx->ndisplay == 0;
 
 	if (is_power_of_2(ctx->ndisplay)) {
 		struct display *new_display;
@@ -1693,10 +1873,11 @@ static int add_display(struct context *ctx, Display *dpy)
 	if (XineramaQueryExtension(dpy, &display->xinerama_event, &display->xinerama_error))
 		display->xinerama_active = XineramaIsActive(dpy);
 
-	display->invisible_cursor = display_load_invisible_cursor(display);
-	display->cursor = None;
-
-	display_cursor_move(display, 0, 0, 0);
+	/* first display (source) is slightly special */
+	if (!first_display) {
+		display->invisible_cursor = display_load_invisible_cursor(display);
+		display_cursor_move(display, 0, 0, 0);
+	}
 
 	return ConnectionNumber(dpy);
 }
@@ -1762,7 +1943,7 @@ static int bumblebee_open(struct context *ctx)
 	DBG(("%s query result '%s'\n", __func__, buf));
 
 	if (strncmp(buf, "Value: ", 7))
-		return -ECONNREFUSED;
+		goto err;
 
 	len = 7;
 	while (buf[len] != '\n' && buf[len] != '\0')
@@ -1822,6 +2003,7 @@ static void display_init_randr_hpd(struct display *display)
 	if (!XRRQueryVersion(display->dpy, &major, &minor))
 		return;
 
+	DBG(("%s - randr version %d.%d\n", DisplayString(display->dpy), major, minor));
 	if (major > 1 || (major == 1 && minor >= 2))
 		XRRSelectInput(display->dpy, display->root, RROutputChangeNotifyMask);
 }
@@ -1890,6 +2072,9 @@ static int last_display_add_clones__randr(struct context *ctx)
 		struct clone *clone = add_clone(ctx);
 		RROutput id;
 
+		if (clone == NULL)
+			return -ENOMEM;
+
 		clone->depth = 24;
 		clone->next = display->clone;
 		display->clone = clone;
@@ -1929,6 +2114,13 @@ static int last_display_add_clones__randr(struct context *ctx)
 			return ret;
 		}
 
+
+		if (o->crtc) {
+			DBG(("%s - disabling active output\n", DisplayString(display->dpy)));
+			XRRSetCrtcConfig(display->dpy, res, o->crtc, CurrentTime,
+					0, 0, None, RR_Rotate_0, NULL, 0);
+		}
+
 		XRRFreeOutputInfo(o);
 	}
 	XRRFreeScreenResources(res);
@@ -1951,6 +2143,9 @@ static int last_display_add_clones__xinerama(struct context *ctx)
 		struct clone *clone = add_clone(ctx);
 		RROutput id;
 
+		if (clone == NULL)
+			return -ENOMEM;
+
 		if (xi[n].width == 0 || xi[n].height == 0)
 			continue;
 
@@ -2008,14 +2203,18 @@ static int last_display_add_clones__xinerama(struct context *ctx)
 static int last_display_add_clones__display(struct context *ctx)
 {
 	struct display *display = last_display(ctx);
-	struct clone *clone = add_clone(ctx);
 	Display *dpy = display->dpy;
+	struct clone *clone;
 	Screen *scr;
 	char buf[80];
 	int ret;
 	RROutput id;
 
-	DBG(("%s(%s)\n", __func__, DisplayString(display->dpy)));
+
+	DBG(("%s(%s)\n", __func__, DisplayString(dpy)));
+	clone = add_clone(ctx);
+	if (clone == NULL)
+		return -ENOMEM;
 
 	clone->depth = 24;
 	clone->next = display->clone;
@@ -2044,7 +2243,7 @@ static int last_display_add_clones__display(struct context *ctx)
 	ret = clone_init_depth(clone);
 	if (ret) {
 		fprintf(stderr, "Failed to negotiate image format for display \"%s\"\n",
-			DisplayString(display->dpy));
+			DisplayString(dpy));
 		return ret;
 	}
 
@@ -2058,7 +2257,7 @@ static int last_display_add_clones__display(struct context *ctx)
 	ret = clone_update_modes__fixed(clone);
 	if (ret) {
 		fprintf(stderr, "Failed to clone display \"%s\"\n",
-			DisplayString(display->dpy));
+			DisplayString(dpy));
 		return ret;
 	}
 
@@ -2072,6 +2271,10 @@ static int last_display_add_clones(struct context *ctx)
 {
 	struct display *display = last_display(ctx);
 
+	display->width = DisplayWidth(display->dpy, DefaultScreen(display->dpy));
+	display->height = DisplayHeight(display->dpy, DefaultScreen(display->dpy));
+	DBG(("%s - initial size %dx%d\n", DisplayString(display->dpy), display->width, display->height));
+
 	if (display->rr_active)
 		return last_display_add_clones__randr(ctx);
 
@@ -2316,24 +2519,42 @@ static void display_flush_send(struct display *display)
 		return;
 	}
 
+	memset(&e, 0, sizeof(e));
 	e.type = display->shm_event;
 	e.send_event = 1;
-	e.serial = 0;
 	e.drawable = display->root;
 	e.major_code = display->shm_opcode;
 	e.minor_code = X_ShmPutImage;
-	e.shmseg = 0;
-	e.offset = 0;
 
 	XSendEvent(display->dpy, display->root, False, 0, (XEvent *)&e);
 	display_mark_flush(display);
 }
 
+static void display_sync(struct display *display)
+{
+	if (display->skip_clone == 0)
+		return;
+
+	if (display->skip_frame++ < 2)
+		return;
+
+	DBG(("%s forcing sync\n", DisplayString(display->dpy)));
+	XSync(display->dpy, False);
+
+	display->flush = 0;
+	display->send = 0;
+
+	/* Event tracking proven unreliable, disable */
+	display->shm_event = 0;
+}
+
 static void display_flush(struct display *display)
 {
 	display_flush_cursor(display);
 	display_flush_send(display);
 
+	display_sync(display);
+
 	if (!display->flush)
 		return;
 
@@ -2590,7 +2811,7 @@ int main(int argc, char **argv)
 		int reconfigure = 0;
 		int rr_update = 0;
 
-		DBG(("polling - enable timer? %d, nfd=%d\n", ctx.timer_active, ctx.nfd));
+		DBG(("polling - enable timer? %d, nfd=%d, ndisplay=%d\n", ctx.timer_active, ctx.nfd, ctx.ndisplay));
 		ret = poll(ctx.pfd + !ctx.timer_active, ctx.nfd - !ctx.timer_active, -1);
 		if (ret <= 0)
 			break;
@@ -2598,7 +2819,7 @@ int main(int argc, char **argv)
 		/* pfd[0] is the timer, pfd[1] is the local display, pfd[2] is the mouse, pfd[3+] are the remotes */
 
 		DBG(("poll reports %d fd awake\n", ret));
-		if (ctx.pfd[1].revents) {
+		if (ctx.pfd[1].revents || XPending(ctx.display[0].dpy)) {
 			DBG(("%s woken up\n", DisplayString(ctx.display[0].dpy)));
 			do {
 				XNextEvent(ctx.display->dpy, &e);
@@ -2656,13 +2877,11 @@ int main(int argc, char **argv)
 				} else {
 					DBG(("unknown event %d\n", e.type));
 				}
-			} while (XPending(ctx.display->dpy) || poll(&ctx.pfd[1], 1, 0) > 0);
-
-			ret--;
+			} while (XEventsQueued(ctx.display->dpy, QueuedAfterReading));
 		}
 
-		for (i = 1; ret && i < ctx.ndisplay; i++) {
-			if (ctx.pfd[i+2].revents == 0)
+		for (i = 1; i < ctx.ndisplay; i++) {
+			if (ctx.pfd[i+2].revents == 0 && !XPending(ctx.display[i].dpy))
 				continue;
 
 			DBG(("%s woken up\n", DisplayString(ctx.display[i].dpy)));
@@ -2672,24 +2891,20 @@ int main(int argc, char **argv)
 				DBG(("%s received event %d\n", DisplayString(ctx.display[i].dpy), e.type));
 				if (ctx.display[i].rr_active && e.type == ctx.display[i].rr_event + RRNotify) {
 					XRRNotifyEvent *re = (XRRNotifyEvent *)&e;
+
+					DBG(("%s received RRNotify, type %d\n", DisplayString(ctx.display[i].dpy), re->subtype));
 					if (re->subtype == RRNotify_OutputChange) {
 						XRROutputPropertyNotifyEvent *ro = (XRROutputPropertyNotifyEvent *)re;
 						struct clone *clone;
 
+						DBG(("%s RRNotify_OutputChange, timestamp %ld\n", DisplayString(ctx.display[i].dpy), ro->timestamp));
 						for (clone = ctx.display[i].clone; clone; clone = clone->next) {
 							if (clone->dst.rr_output == ro->output)
 								rr_update = clone->rr_update = 1;
 						}
 					}
 				}
-			} while (XPending(ctx.display[i].dpy) || poll(&ctx.pfd[i+2], 1, 0) > 0);
-
-			ret--;
-		}
-
-		if (reconfigure) {
-			context_update(&ctx);
-			display_reset_damage(ctx.display);
+			} while (XEventsQueued(ctx.display[i].dpy, QueuedAfterReading));
 		}
 
 		if (rr_update) {
@@ -2697,6 +2912,9 @@ int main(int argc, char **argv)
 				clone_update(&ctx.clones[i]);
 		}
 
+		if (reconfigure && context_update(&ctx))
+			display_reset_damage(ctx.display);
+
 		XPending(ctx.record);
 
 		if (ctx.timer_active && read(ctx.timer, &count, sizeof(count)) > 0) {
